/* -*- mode: c++; c-basic-offset: 4; -*- */
#include <frame_source.hh>
#include "ParseLine.hh"
#include "driver_table.hh"
#include "scandir.hh"
#include "gpstime.hh"
#include <stdexcept>
#include <iostream>

using namespace sends;
using namespace std;

//======================================  Constructor
frame_source::frame_source(void) 
    : mDebug(0)
{
}

//======================================  Destructor
frame_source::~frame_source(void) {
}

//======================================  Dump all maps
std::ostream& 
frame_source::dump(std::ostream& out, int lvl) const {
    out << endl;
    out << "Online frame sources" << endl;
    dump_list(mOnlineMap, out, lvl);
    out << endl;
    out << "Cache frame sources" << endl;
    dump_list(mCacheMap, out, lvl);
    out << endl;
    out << "Archive frame sources" << endl;
    dump_list(mArchiveMap, out, lvl);
    out << endl;
    out << "Offline frame sources" << endl;
    dump_list(mOfflineMap, out, lvl);
    return out;
}

//======================================  Dump the specified map
std::ostream& 
frame_source::dump_list(const frame_map& frm, std::ostream& out, int lvl) const
{
    for (const_frmap_iter i=frm.begin(); i != frm.end(); ++i) {
	i->second.read_lock();
	long N=i->second.size();
	switch (lvl) {
	case 0:
	    break;
	case 1: {
	    out << i->first;
	    gps_type start(0), stop(0), nCol(0);
	    for (long j=0; j<N; ++j) {
		if (!start) {
		    start = i->second[j].start();
		    stop  = i->second[j].stop();
		} else if (stop == i->second[j].start()) {
		    stop = i->second[j].stop();
		} else {
		    if (!(nCol%3) && nCol) out << endl << "          ";
		    out << "  " << start << "-" << stop;
		    nCol++;
		    start = i->second[j].start();
		    stop  = i->second[j].stop();
		}
	    }
	    if (start) {
		if (!(nCol%3) && nCol) out << endl << "          ";
		out << "  " << start << "-" << stop;
	    }
	    out << endl;
	    break;
	}
	case 2:
	    for (long j=0; j<N; ++j) {
		i->second[j].dump(out);
	    }
	    break;
	default:
	    break;
	}
	i->second.unlock();
    }
    return out;
}

#ifdef LOC_LIST_LOCKED_FINDS
//======================================  Find a  {type, gps}
const frame_loc* 
frame_source::find(const string& type, gps_type gps) const {
    const frame_loc* fl = find(loc_list::kOnline,  type, gps);
    if (!fl)   fl = find(loc_list::kCache,   type, gps);
    if (!fl)   fl = find(loc_list::kArchive, type, gps);
    if (!fl)   fl = find(loc_list::kOffline, type, gps);
    return fl;
}
#endif

//======================================  Find a  {type, gps}
bool
frame_source::find(const string& type, gps_type gps, frame_loc& fl) const {
    if (find(loc_list::kOnline,  type, gps, fl)) return true;
    if (find(loc_list::kCache,   type, gps, fl)) return true;
    if (find(loc_list::kArchive, type, gps, fl)) return true;
    if (find(loc_list::kOffline, type, gps, fl)) return true;
    return false;
}

#ifdef LOC_LIST_LOCKED_FINDS
//======================================  Find a  {type, gps}
const frame_loc* 
frame_source::find(arch_type al, const string& type, gps_type gps) const {
    const loc_list* plist = list_ptr(al, type);
    if (!plist) return 0;
    return plist->find_loc(gps);
}

//======================================  Find a  {type, gps}
frame_loc*
frame_source::find(arch_type al, const string& type, gps_type gps) {
    loc_list* plist = list_ptr(al, type);
    if (!plist) return 0;
    return plist->find_loc(gps);
}
#endif

//======================================  Find a  {type, gps}
bool
frame_source::find(arch_type al, const string& type, gps_type gps,
		   frame_loc& fl) const {
    const loc_list* plist = list_ptr(al, type);
    if (!plist) return false;
    return plist->find_loc(gps, fl);
}

#ifdef LOC_LIST_LOCKED_FINDS
//======================================  Find a  {type, gps}
frame_loc*
frame_source::find_back(arch_type al, const string& type) {
    loc_list* plist = list_ptr(al, type);
    if (!plist || plist->empty()) return 0;
    plist->write_lock();
    return &(plist->back());
}

//======================================  Find a constant {type, gps}
const frame_loc*
frame_source::find_back(arch_type al, const string& type) const {
    const loc_list* plist = list_ptr(al, type);
    if (!plist || plist->empty()) return 0;
    plist->read_lock();
    return &(plist->back());
}
#endif

//======================================  Find a  {type, gps}
bool
frame_source::find_back(arch_type al, const string& type, frame_loc& fl) const {
    const loc_list* plist = list_ptr(al, type);
    if (!plist) return false;
    return plist->find_back(fl);
}

#ifdef LOC_LIST_LOCKED_FINDS
//======================================  Find the first location for the type
const frame_loc*
frame_source::find_first(arch_type a, const std::string& pfx) const {
    const loc_list* plist= list_ptr(a, pfx);
    if (!plist)  return 0;
    plist->read_lock();
    if (!plist->empty()) return &*(plist->begin());
    plist->unlock();
    return NULL;
}
#endif

//======================================  Find the first location for the type
bool
frame_source::find_first(arch_type a, const std::string& pfx,
			 frame_loc& fl) const {
    const loc_list* plist = list_ptr(a, pfx);
    if (!plist)  return false;
    return plist->find_front(fl);
}

//======================================  Find the first location for the type
const loc_list*
frame_source::find_first(const std::string& pfx) const {
    const loc_list* r  = list_ptr(loc_list::kOnline,  pfx);
    if (r && r->empty()) r = 0;

    const loc_list* ll = list_ptr(loc_list::kCache,   pfx);
    if (ll && ll->empty()) ll = 0;
    if (!r || (ll != 0 && ll->front().start() < r->front().start())) {
	r = ll;
    }

    ll = list_ptr(loc_list::kArchive, pfx);
    if (ll && ll->empty()) ll = 0;
    if (!r || (ll != 0 && ll->front().start() < r->front().start())) {
	r = ll;
    }

    ll = list_ptr(loc_list::kOffline, pfx);
    if (!r || (ll != 0 && ll->front().start() < r->front().start())) {
	r = ll;
    }
    return r;
}

//======================================  Find gap around gps
void 
frame_source::find_gap(const std::string& pfx, arch_type a, gps_type gps, 
		       gpsseg_type& gap) const {
    const loc_list* plist = list_ptr(a, pfx);
    gps_type gap_max = inf_gps; 
    gps_type gap_min = 0; 
    if (plist && !plist->empty()) {
	plist->read_lock();
	unsigned long inx = plist->upper_bound(gps);
	if (inx < plist->size()) {
	    if (gps < (*plist)[inx].start()) {
		gap_max = (*plist)[inx].start();
	    } else {
		gap_max = gps;
	    }
	}
	if (inx--) gap_min = (*plist)[inx].stop();
	plist->unlock();
    }
    gap = gpsseg_type(gap_min, gap_max);
}

//======================================  Find gap around gps
void
frame_source::find_gap(const string& pfx, gps_type gps, 
		       gpsseg_type& gap) const {
    find_gap(pfx, loc_list::kCache, gps, gap);
    if (gap.first == gap.second) return;
    gpsseg_type temp;
    find_gap(pfx, loc_list::kArchive, gps, temp);
    if (temp.first  > gap.first ) gap.first  = temp.first;
    if (temp.second < gap.second) gap.second = temp.second;
    if (gap.first >= gap.second) {
	gap = gpsseg_type(gps, gps);
    }
}

//======================================  Find the first location for the type
const loc_list*
frame_source::find_last(const std::string& pfx) const {
    const loc_list* r  = list_ptr(loc_list::kOnline,  pfx);
    if (r && r->empty()) r = 0;

    const loc_list* ll = list_ptr(loc_list::kCache,   pfx);
    if (ll && ll->empty()) ll = 0;
    if (!r || (ll != 0 && ll->back().stop() > r->front().stop())) {
	r = ll;
    }

    ll = list_ptr(loc_list::kArchive, pfx);
    if (ll && ll->empty()) ll = 0;
    if (!r || (ll != 0 && ll->back().stop() > r->back().stop())) {
	r = ll;
    }

    ll = list_ptr(loc_list::kOffline, pfx);
    if (!r || (ll != 0 && ll->back().stop() > r->back().stop())) {
	r = ll;
    }
    return r;
}

//======================================  Get the frame length
gps_type
frame_source::frame_length(const std::string& type, gps_type gps) const {
    frame_loc fl;
    if (!find(type, gps, fl)) return 0;
    return fl.duration();
}

//======================================  Insert a frame location.
void 
frame_source::insert(arch_type arch, const frame_loc& loc) {
    //----------------------------------  Find the type vector. Insert an
    //                                    empty vector if none found.
    loc_list* plist = list_ptr(arch, loc.prefix());

    if (!plist) {
	ref_map(arch).insert(frmap_node(loc.prefix(), loc_list()));
	plist = list_ptr(arch, loc.prefix());
	plist->set_archive(arch);
    }

    //-----------------------------------  Insert the frame_loc into the vector.
    plist->write_lock();
    plist->insert(loc);
    plist->unlock();
}

//======================================  Get a pointer to the specified list.
const loc_list* 
frame_source::list_ptr(arch_type arch, const std::string& pfx) const {
    const frame_map& map = ref_map(arch);
    const_frmap_iter pnode = map.find(pfx);
    if (pnode == map.end()) return 0;
    return &(pnode->second);
}

loc_list* 
frame_source::list_ptr(arch_type arch, const std::string& pfx) {
    frame_map& map = ref_map(arch);
    frmap_iter pnode = map.find(pfx);
    if (pnode == map.end()) return 0;
    return &(pnode->second);
}

const loc_list* 
frame_source::list_ptr(const string& pfx, gps_type gps) const{
    //----------------------------------  Handle online data separately.
    if (!gps) {
	return list_ptr(loc_list::kOnline, pfx);
    }

    //----------------------------------  Handle offline archive types.
    else {
	const loc_list* lp = list_ptr(loc_list::kCache, pfx);
	if (lp && lp->in_range(gps)) return lp;
	lp = list_ptr(loc_list::kArchive, pfx);
	if (lp && lp->in_range(gps)) return lp;
	lp = list_ptr(loc_list::kOffline, pfx);
	if (lp && lp->in_range(gps)) return lp;
    }
    return 0;
}

//======================================  calculate memory usage
index_type
frame_source::mem_use(void) const {
    index_type total = 4 * sizeof(frame_map);
    for (const_frmap_iter i=mOnlineMap.begin(); i != mOnlineMap.end(); i++) {
	total += sizeof(frmap_node) + i->first.size() + i->second.mem_use()
	    - sizeof(loc_list);
    }
    for (const_frmap_iter i=mCacheMap.begin(); i != mCacheMap.end(); i++) {
	total += sizeof(frmap_node) + i->first.size() + i->second.mem_use()
	    - sizeof(loc_list);
    }
    for (const_frmap_iter i=mArchiveMap.begin(); i != mArchiveMap.end(); i++) {
	total += sizeof(frmap_node) + i->first.size() + i->second.mem_use()
	    - sizeof(loc_list);
    }
    for (const_frmap_iter i=mOfflineMap.begin(); i != mOfflineMap.end(); i++) {
	total += sizeof(frmap_node) + i->first.size() + i->second.mem_use()
	    - sizeof(loc_list);
    }
    return total;
}

//======================================  Add an entry to a range_vect
inline void
add_rv(frame_source::range_vect& rv, const std::string& ft, const epoch& epik) {
    for (size_t i=0; i != rv.size(); i++) {
	if (rv[i].first == ft) {
	    rv[i].second.merge(epik);
	    return;
	}
    }
    rv.push_back(frame_source::range_vect::value_type(ft, epik));
}

//======================================  List source type validity ranges
void
frame_source::range_list(range_vect& rv) const {
    for (const_frmap_iter i=mCacheMap.begin(); i != mCacheMap.end(); i++) {	
	add_rv(rv, i->first, i->second.range());
    }
    for (const_frmap_iter i=mArchiveMap.begin(); i != mArchiveMap.end(); i++) {	
	add_rv(rv, i->first, i->second.range());
    }
    for (const_frmap_iter i=mOfflineMap.begin(); i != mOfflineMap.end(); i++) {	
	add_rv(rv, i->first, i->second.range());
    }
}

//======================================  Read a file
void 
frame_source::read(const std::string& file) {
    ParseLine pl(file.c_str());
    if (!pl.isOpen()) {
	throw runtime_error(string("unable to open file: ") + file);
    }
    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (nArg < 7) {
	    cerr << "Line: " << pl.getLineNumber() << " of file: " << file 
		 << " is not a valid source specifier." << endl; 
	    throw runtime_error("Input line too short");
	}
	string   type  = pl[0];
	gps_type start = pl.getInt(1);
	gps_type stop  = pl.getInt(2);
	gps_type dt    = pl.getInt(3);
	arch_type arch = loc_list::get_arch_type(pl[4]);

	frame_loc::driver_type driver = driver_table::default_driver;
	string drivest = pl[5];
	if (drivest != "-") {
	    driver = TheDriverTable.get_type(drivest);
	}

	if (arch == loc_list::kOnline && !TheDriverTable[driver].online()) {
	    cerr << "*** Error in frame_source:read: "
		 << "Online data requested from offline driver: " 
		 << TheDriverTable[driver].name() << endl;
	}

	string path = pl[6];

	//------------------------------  Scan the file system if dt is zero
	if (!dt) {
	    cout << "Scan frames in: " << path << endl;
	    loc_list vec;
	    vec.set_archive(arch);
	    gps_type tStart = gps_now();
	    scan(path, type, vec, driver);
	    vec.sort();
	    vec.coalesce();
	    index_type N = vec.size();
	    cout << "Number of frame location segments is: " << N << endl;
	    for (index_type i=0; i<N; ++i) {
		if (start && vec[i].stop()  < start) continue;
		if (stop  && vec[i].start() >= stop) continue;
		insert(arch, vec[i]);
	    }

	    //--------------------------  Set scan time if data found.
	    loc_list* lptr = list_ptr(arch, type);
	    if (lptr) {
		lptr->set_scan_time(tStart);
		lptr->set_scan_path(path);
	    }
	}

	//------------------------------  Otherwise insert it directly
	else {
	    insert(arch, frame_loc(path, type, start, stop, dt, driver));
	}
    }
}

//======================================  Reference to specified map
frame_source::frame_map& 
frame_source::ref_map(loc_list::arch_type arch) {
    switch (arch) {
    case loc_list::kOnline:
	return mOnlineMap;
    case loc_list::kCache:
	return mCacheMap;
    case loc_list::kArchive:
	return mArchiveMap;
    case loc_list::kOffline:
	return mOfflineMap;
    }
    throw runtime_error("Undefined archive type.");
}

//======================================  Constant reference to specified map
const frame_source::frame_map& 
frame_source::ref_map(loc_list::arch_type arch) const {
    switch (arch) {
    case loc_list::kOnline:
	return mOnlineMap;
    case loc_list::kCache:
	return mCacheMap;
    case loc_list::kArchive:
	return mArchiveMap;
    case loc_list::kOffline:
	return mOfflineMap;
    }
    throw runtime_error("Undefined archive type.");
}

//======================================  Reference to specified map
void 
frame_source::rescan(loc_list::arch_type arch, const std::string& prefix) {
    loc_list* plist = list_ptr(arch, prefix);
    if (!plist) return;
    string path = plist->scan_path();
    if (path.empty()) {
	string msg = "frame_source::rescan called for unscanned list: ";
	msg += string(loc_list::arch_name(arch)) + ":" + prefix;
	throw runtime_error(msg);
    }

    frame_loc::driver_type dinx = driver_table::default_driver;
    if (!plist->empty()) dinx = plist->front().driver();
    loc_list vec;
    gps_type tStart = gps_now();
    scan(path, prefix, vec, dinx);
    vec.coalesce();
    plist->set_scan_time(tStart);
    plist->swap_vec(vec);
}

//======================================  Inline digit test.
inline bool
test_digit(char c) {
    if (c < '0') return false;
    if (c > '9') return false;
    return true;
}

//======================================  Test name against pattern
static bool
test_name(const std::string& pattern, const std::string& name) {
    string::size_type inxname = 0;
    string::size_type lname   = name.size();

    string::size_type N = pattern.size();
    for (string::size_type ipat=0; ipat<N; ++ipat) {
	//------------------------------  OK if the char equals the pattern. 
	if (pattern[ipat] == name[inxname]) {
	    inxname++;
	}

	//------------------------------  Check for an escape char;
	else if (pattern[ipat] == '%') {
	    ipat++;
	    string::size_type wid = 0;
	    while (test_digit(pattern[ipat]) ) {
		wid *= 10;
		wid += pattern[ipat++] - '0';
	    }

	    switch (pattern[ipat]) {

	    //--------------------------  width has a different meaning for %r
	    case 'r':
		wid = 0;

	    //--------------------------  Numeric field 
	    case 'D':
	    case 'd':
	    case 'Y':
	    case 'M':
		//----------------------  Constant field length
		if (wid) {
		    if (inxname + wid > lname) return false;
		    for (string::size_type i=0; i<wid; ++i) {
			if (!test_digit(name[inxname++])) return false;
		    }
		}

		//----------------------  Variable field length
		else {
		    if (inxname >= lname) return false;
		    while (inxname < lname && test_digit(name[inxname])) {
			inxname++;
		    }
		}
	        break;

	    //--------------------------  Not a known escape
	    default:
		return false;
	    }
	}
	else {
	    return false; 
	}
    } 

    return (inxname == lname);
}

//======================================  Scan for frame files
void 
frame_source::scan(const string& location, const string& prefix,
		   loc_list& vec, frame_loc::driver_type drive) const {

    //----------------------------------  Look for a formatted directory name
    string::size_type fmtc = location.find_first_of('%');
    if (fmtc != string::npos) {
	string::size_type dend = location.find_last_of('/', fmtc);
	string::size_type spat = dend + 1;
	string basedir;
	if (dend == 0) {
	    basedir = "/";
	}
	else if (dend == string::npos) {
	    basedir = "./";
	    spat    = 0;
	}
	else {
	    basedir = location.substr(0, dend);
	}

	//------------------------------  Open the directory
	string::size_type epat = location.find_first_of('/', spat);
	scandir dirp(basedir);
	if (!dirp.is_open()) return;

	//------------------------------  scan the directory for matching
	string pat = location.substr(spat, epat-spat);
	index_type vindex = vec.size();
	while (dirp.next_entry()) {
	    if (dirp.is_dir() && test_name(pat, dirp.file_name())) {
		string newloc = location;
		newloc.replace(spat, epat-spat, dirp.file_name());
		if (mDebug > 2) cout << "Recursive scan: " << newloc 
				     << " for: " << prefix << endl;
		scan(newloc, prefix, vec, drive);
	    }
	}

	index_type N = vec.size();
	for (index_type inx=vindex; inx<N; ++inx) {
	    vec[inx].set_path(location);
	}
	vec.coalesce();
    }

    //----------------------------------  Scan unambiguous directory
    else {
	scandir dirp(location);
	if (!dirp.is_open()) return;

	//------------------------------  Scan the directory for frame files
	loc_list dvec;
	while (dirp.next_entry()) {
	    if (mDebug > 3) cout << "Test path: " << dirp.file_path() << " - ";
	    if (dirp.is_frame()) {
		gps_type gps = dirp.gps();
		gps_type dt  = dirp.duration();
		frame_loc fl(location, prefix, gps, gps+dt, dt, drive);
		if (fl.file_path(gps) == dirp.file_path()) {
		    loc_list::iterator itr = dvec.lower_bound(fl);
		    if (itr != dvec.end() && itr->merge(fl)) {
			continue;
		    } 
		    else if (itr != dvec.begin() && (--itr)->merge(fl)) {
			continue;
		    } 
		    else {
			dvec.insert(fl);
		    }
		}
		else if (mDebug > 3) {
		    cout << "Pattern mismatch." << endl;
		}
	    }
	    else if (mDebug > 3) {
		cout << "Not a regular file" << endl;
	    }
	}

	//------------------------------  Merge new segments into vector.
	vec.merge(dvec);
    }
}
